import logging
from typing import Dict, Tuple

import pandas as pd
from sklearn.ensemble import RandomForestRegressor
from sklearn.metrics import r2_score
from sklearn.model_selection import train_test_split

{% if cookiecutter.enable_mlflow_integration|lower != "false" %}
from .. import mlflow_helpers
{% endif %}

def split_data(data: pd.DataFrame, parameters: Dict) -> Tuple:
    """Splits data into features and targets training and test sets.

    Args:
        data: Data containing features and target.
        parameters: Parameters defined in parameters/data_science.yml.
    Returns:
        Split data.
    """
    X = data[parameters["features"]]
    y = data["price"]
    X_train, X_test, y_train, y_test = train_test_split(
        X, y, test_size=parameters["test_size"], random_state=parameters["random_state"]
    )
    return X_train, X_test, y_train, y_test


def train_model(
    X_train: pd.DataFrame, y_train: pd.Series, parameters: dict
) -> RandomForestRegressor:
    """Trains the regression model.

    Args:
        X_train: Training data of independent features.
        y_train: Training data for price.

    Returns:
        Trained model.
    """
    regressor = RandomForestRegressor(random_state=parameters["random_state"])
{% if cookiecutter.enable_mlflow_integration|lower != "false" %}
    with mlflow_helpers.log_model() as run:
        regressor.fit(X_train, y_train)
{% else %}
    regressor.fit(X_train, y_train)
{% endif %}
    return regressor


def evaluate_model(
    regressor: RandomForestRegressor, X_test: pd.DataFrame, y_test: pd.Series
):
    """Calculates and logs the coefficient of determination.

    Args:
        regressor: Trained model.
        X_test: Testing data of independent features.
        y_test: Testing data for price.
    """
    y_pred = regressor.predict(X_test)
    score = r2_score(y_test, y_pred)
    logger = logging.getLogger(__name__)
    logger.info("Model has a coefficient R^2 of %.3f on test data.", score)

{% if cookiecutter.enable_mlflow_integration|lower != "false" %}
    mlflow_helpers.log_metric("r2", score)
    mlflow_helpers.log_parameter("snowflake_warehouse",
                                 mlflow_helpers.get_current_warehouse())
    mlflow_helpers.log_parameter("snowflake_account",
                                 mlflow_helpers.get_snowflake_account())
{% endif %}
